import re
import sys
from spec_tests import get_tests, do_test
from cmark import CMark
import argparse

parser = argparse.ArgumentParser(description='Run cmark roundtrip tests.')
parser.add_argument('-p', '--program', dest='program', nargs='?', default=None,
        help='program to test')
parser.add_argument('-s', '--spec', dest='spec', nargs='?', default='spec.txt',
        help='path to spec')
parser.add_argument('-P', '--pattern', dest='pattern', nargs='?',
        default=None, help='limit to sections matching regex pattern')
parser.add_argument('--library-dir', dest='library_dir', nargs='?',
        default=None, help='directory containing dynamic library')
parser.add_argument('--no-normalize', dest='normalize',
        action='store_const', const=False, default=True,
        help='do not normalize HTML')
parser.add_argument('-n', '--number', type=int, default=None,
        help='only consider the test with the given number')
args = parser.parse_args(sys.argv[1:])

spec = sys.argv[1]

def converter(md):
  cmark = CMark(prog=args.program, library_dir=args.library_dir)
  [ec, result, err] = cmark.to_commonmark(md)
  if ec == 0:
    [ec, html, err] = cmark.to_html(result)
    if ec == 0:
        # In the commonmark writer we insert dummy HTML
        # comments between lists, and between lists and code
        # blocks.  Strip these out, since the spec uses
        # two blank lines instead:
        return [ec, re.sub('<!-- end list -->\n', '', html), '']
    else:
        return [ec, html, err]
  else:
    return [ec, result, err]

tests = get_tests(args.spec)
result_counts = {'pass': 0, 'fail': 0, 'error': 0, 'skip': 0}
for test in tests:
    do_test(converter, test, args.normalize, result_counts)

exit(result_counts['fail'] + result_counts['error'])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ctypes import *
from subprocess import Popen, PIPE
import platform
import os

class cmark_mem(Structure):
    _fields_ = [("calloc", c_void_p),
                ("realloc", c_void_p),
                ("free", CFUNCTYPE(None, c_void_p))]

def pipe_through_prog(prog, text):
    p1 = Popen(prog.split(), stdout=PIPE, stdin=PIPE, stderr=PIPE)
    [result, err] = p1.communicate(input=text.encode('utf-8'))
    return [p1.returncode, result.decode('utf-8'), err]

def to_html(lib, text):
    get_alloc = lib.cmark_get_default_mem_allocator
    get_alloc.restype = POINTER(cmark_mem)
    free_func = get_alloc().contents.free

    markdown = lib.cmark_markdown_to_html
    markdown.restype = POINTER(c_char)
    markdown.argtypes = [c_char_p, c_size_t, c_int]

    textbytes = text.encode('utf-8')
    textlen = len(textbytes)
    # 1 << 17 == CMARK_OPT_UNSAFE
    cstring = markdown(textbytes, textlen, 1 << 17)
    result = string_at(cstring).decode('utf-8')
    free_func(cstring)

    return [0, result, '']

def to_commonmark(lib, text):
    get_alloc = lib.cmark_get_default_mem_allocator
    get_alloc.restype = POINTER(cmark_mem)
    free_func = get_alloc().contents.free

    parse_document = lib.cmark_parse_document
    parse_document.restype = c_void_p
    parse_document.argtypes = [c_char_p, c_size_t, c_int]

    render_commonmark = lib.cmark_render_commonmark
    render_commonmark.restype = POINTER(c_char)
    render_commonmark.argtypes = [c_void_p, c_int, c_int]

    free_node = lib.cmark_node_free
    free_node.argtypes = [c_void_p]

    textbytes = text.encode('utf-8')
    textlen = len(textbytes)
    node = parse_document(textbytes, textlen, 0)
    cstring = render_commonmark(node, 0, 0)
    result = string_at(cstring).decode('utf-8')
    free_func(cstring)
    free_node(node)

    return [0, result, '']

class CMark:
    def __init__(self, prog=None, library_dir=None):
        self.prog = prog
        if prog:
            prog += ' --unsafe'
            self.to_html = lambda x: pipe_through_prog(prog, x)
            self.to_commonmark = lambda x: pipe_through_prog(prog + ' -t commonmark', x)
        else:
            sysname = platform.system()
            if sysname == 'Darwin':
                libnames = [ "libcmark.dylib" ]
            elif sysname == 'Windows':
                libnames = [ "cmark.dll", "libcmark.dll" ]
            else:
                libnames = [ "libcmark.so" ]
            if not library_dir:
                library_dir = os.path.join("build", "src")
            for libname in libnames:
                candidate = os.path.join(library_dir, libname)
                if os.path.isfile(candidate):
                    libpath = candidate
                    break
            cmark = CDLL(libpath)
            self.to_html = lambda x: to_html(cmark, x)
            self.to_commonmark = lambda x: to_commonmark(cmark, x)




#!/usr/bin/env python

# Creates a man page from a C file.

# Comments beginning with `/**` are treated as Groff man, except that
# 'this' is converted to \fIthis\f[], and ''this'' to \fBthis\f[].

# Non-blank lines immediately following a man page comment are treated
# as function signatures or examples and parsed into .Ft, .Fo, .Fa, .Fc. The
# immediately preceding man documentation chunk is printed after the example
# as a comment on it.

# That's about it!

import sys, re, os, platform
from datetime import date
from ctypes import CDLL, c_char_p, c_long, c_void_p

sysname = platform.system()

curdir = os.getcwd()

if sysname == 'Darwin':
    cmark = CDLL(curdir + "/build/src/libcmark.dylib")
else:
    cmark = CDLL(curdir + "/build/src/libcmark.so")

parse_document = cmark.cmark_parse_document
parse_document.restype = c_void_p
parse_document.argtypes = [c_char_p, c_long]

render_man = cmark.cmark_render_man
render_man.restype = c_char_p
render_man.argtypes = [c_void_p, c_long, c_long]

cmark_version_string = cmark.cmark_version_string
cmark_version_string.restype = c_char_p
cmark_version_string.argtypes = []

def md2man(text):
    if sys.version_info >= (3,0):
        textbytes = text.encode('utf-8')
        textlen = len(textbytes)
        return render_man(parse_document(textbytes, textlen), 0, 72).decode('utf-8')
    else:
        textbytes = text
        textlen = len(text)
        return render_man(parse_document(textbytes, textlen), 0, 72)

comment_start_re = re.compile(r'^\/\*\* ?')
comment_delim_re = re.compile(r'^[/ ]\** ?')
comment_end_re = re.compile(r'^ \**\/')
function_re = re.compile(r'^ *(?:CMARK_EXPORT\s+)?(?P<type>(?:const\s+)?\w+(?:\s*[*])?)\s*(?P<name>\w+)\s*\((?P<args>[^)]*)\)')
blank_re = re.compile(r'^\s*$')
macro_re = re.compile(r'CMARK_EXPORT *')
typedef_start_re = re.compile(r'typedef.*{$')
typedef_end_re = re.compile(r'}')
single_quote_re = re.compile(r"(?<!\w)'([^']+)'(?!\w)")
double_quote_re = re.compile(r"(?<!\w)''([^']+)''(?!\w)")

def handle_quotes(s):
    return re.sub(double_quote_re, r'**\g<1>**', re.sub(single_quote_re, r'*\g<1>*', s))

typedef = False
mdlines = []
chunk = []
sig = []

if len(sys.argv) > 1:
    sourcefile = sys.argv[1]
else:
    print("Usage:  make_man_page.py sourcefile")
    exit(1)

with open(sourcefile, 'r') as cmarkh:
    state = 'default'
    for line in cmarkh:
        # state transition
        oldstate = state
        if comment_start_re.match(line):
            state = 'man'
        elif comment_end_re.match(line) and state == 'man':
            continue
        elif comment_delim_re.match(line) and state == 'man':
            state = 'man'
        elif not typedef and blank_re.match(line):
            state = 'default'
        elif typedef and typedef_end_re.match(line):
            typedef = False
        elif typedef_start_re.match(line):
            typedef = True
            state = 'signature'
        elif state == 'man':
            state = 'signature'

        # handle line
        if state == 'man':
            chunk.append(handle_quotes(re.sub(comment_delim_re, '', line)))
        elif state == 'signature':
            ln = re.sub(macro_re, '', line)
            if typedef or not re.match(blank_re, ln):
                sig.append(ln)
        elif oldstate == 'signature' and state != 'signature':
            if len(mdlines) > 0 and mdlines[-1] != '\n':
                mdlines.append('\n')
            rawsig = ''.join(sig)
            m = function_re.match(rawsig)
            mdlines.append('.PP\n')
            if m:
                mdlines.append('\\fI' + m.group('type') + '\\f[]' + ' ')
                mdlines.append('\\fB' + m.group('name') + '\\f[]' + '(')
                first = True
                for argument in re.split(',', m.group('args')):
                    if not first:
                        mdlines.append(', ')
                    first = False
                    mdlines.append('\\fI' + argument.strip() + '\\f[]')
                mdlines.append(')\n')
            else:
                mdlines.append('.nf\n\\fC\n.RS 0n\n')
                mdlines += sig
                mdlines.append('.RE\n\\f[]\n.fi\n')
            if len(mdlines) > 0 and mdlines[-1] != '\n':
                mdlines.append('\n')
            mdlines += md2man(''.join(chunk))
            mdlines.append('\n')
            chunk = []
            sig = []
        elif oldstate == 'man' and state != 'signature':
            if len(mdlines) > 0 and mdlines[-1] != '\n':
                mdlines.append('\n')
            mdlines += md2man(''.join(chunk)) # add man chunk
            chunk = []
            mdlines.append('\n')

sys.stdout.write('.TH ' + os.path.basename(sourcefile).replace('.h','') + ' 3 "' + date.today().strftime('%B %d, %Y') + '" "cmark ' + cmark_version_string().decode('utf-8') + '" "Library Functions Manual"\n')
sys.stdout.write(''.join(mdlines))



# Creates a C lookup table for Unicode case folding (https://unicode.org/Public/UCD/latest/ucd/CaseFolding.txt).
# Usage: python3 tools/make_case_fold_inc.py < data/CaseFolding.txt > src/case_fold.inc

import sys, re

prog = re.compile('([0-9A-F]+); [CF];((?: [0-9A-F]+)+);')
main_table = []
repl_table = []
repl_idx = 0
test = ''
test_result = ''

for line in sys.stdin:
    m = prog.match(line)
    if m is None:
        continue

    cp = int(m[1], 16);
    if cp < 0x80:
        continue

    repl = b''
    for x in m[2].split():
        repl += chr(int(x, 16)).encode('UTF-8')

    # Generate test case
    if len(main_table) % 20 == 0:
        test += chr(cp)
        test_result += repl.decode('UTF-8')

    # 17 bits for code point
    if cp >= (1 << 17):
        raise Exception("code point too large")

    # 12 bits for upper bits of replacement index
    # The lowest bit is always zero.
    if repl_idx // 2 >= (1 << 12):
        raise Exception("too many replacements")

    # 3 bits for size of replacement
    repl_size = len(repl)
    if repl_size >= (1 << 3):
        raise Exception("too many replacement chars")

    main_table += [ cp | repl_idx // 2 << 17 | repl_size << 29 ]
    repl_table += repl
    repl_idx += repl_size

    # Make sure that repl_idx is even
    if repl_idx % 2 != 0:
        repl_table += [0]
        repl_idx += 1

# Print test case
if False:
    print("test:", test)
    print("test_result:", test_result)
    sys.exit(0)

print("""// Generated by tools/make_case_fold_inc.py

#define CF_MAX            (1 << 17)
#define CF_TABLE_SIZE     %d
#define CF_CODE_POINT(x)  ((x) & 0x1FFFF)
#define CF_REPL_IDX(x)    ((((x) >> 17) & 0xFFF) * 2)
#define CF_REPL_SIZE(x)   ((x) >> 29)

static const uint32_t cf_table[%d] = {""" % (len(main_table), len(main_table)))

i = 0
size = len(main_table)
for value in main_table:
    if i % 6 == 0:
        print("  ", end="")
    print("0x%X" % value, end="")
    i += 1
    if i == size: print()
    elif i % 6 == 0: print(",")
    else: print(", ", end="")

print("""};

static const unsigned char cf_repl[%d] = {""" % len(repl_table))

i = 0
size = len(repl_table)
for value in repl_table:
    if i % 12 == 0:
        print("  ", end="")
    print("0x%02X" % value, end="")
    i += 1
    if i == size: print()
    elif i % 12 == 0: print(",")
    else: print(", ", end="")

print("};")



# Creates C data structures for binary lookup table of entities,
# using python's html5 entity data.
# Usage: python3 tools/make_entities_inc.py > src/entities.inc

import html

entities5 = html.entities.html5

# Remove keys without semicolons. HTML5 allows some named character
# references without a trailing semicolon.
entities = sorted([(k[:-1], entities5[k]) for k in entities5.keys() if k[-1] == ';'])

main_table = []
text_table = b''
text_idx = 0

for (ent, repl) in entities:
    ent_bytes = ent.encode('UTF-8')
    ent_size = len(ent_bytes)
    repl_bytes = repl.encode('UTF-8')
    repl_size = len(repl_bytes)

    if text_idx >= (1 << 15):
        raise Exception("text index too large")
    if ent_size >= (1 << 5):
        raise Exception("entity name too long")
    if repl_size >= (1 << 3):
        raise Exception("entity replacement too long")

    main_table += [ text_idx | ent_size << 15 | repl_size << 20 ]

    text_table += ent_bytes + repl_bytes
    text_idx += ent_size + repl_size

print("""/* Autogenerated by tools/make_headers_inc.py */

#define ENT_MIN_LENGTH      2
#define ENT_MAX_LENGTH      32
#define ENT_TABLE_SIZE      %d
#define ENT_TEXT_IDX(x)     ((x) & 0x7FFF)
#define ENT_NAME_SIZE(x)    (((x) >> 15) & 0x1F)
#define ENT_REPL_SIZE(x)    ((x) >> 20)

static const uint32_t cmark_entities[%d] = {""" % (len(main_table), len(main_table)));

i = 0
size = len(main_table)
for value in main_table:
    if i % 6 == 0:
        print("  ", end="")
    print("0x%X" % value, end="")
    i += 1
    if i == size: print()
    elif i % 6 == 0: print(",")
    else: print(", ", end="")

print("""};

static const unsigned char cmark_entity_text[%d] = {""" % len(text_table))

i = 0
size = len(text_table)
for value in text_table:
    if i % 12 == 0:
        print("  ", end="")
    print("0x%02X" % value, end="")
    i += 1
    if i == size: print()
    elif i % 12 == 0: print(",")
    else: print(", ", end="")

print("};")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re
import argparse
import sys
import platform
import itertools
import multiprocessing
import queue
import time
from cmark import CMark

TIMEOUT = 5

parser = argparse.ArgumentParser(description='Run cmark tests.')
parser.add_argument('--program', dest='program', nargs='?', default=None,
        help='program to test')
parser.add_argument('--library-dir', dest='library_dir', nargs='?',
        default=None, help='directory containing dynamic library')
args = parser.parse_args(sys.argv[1:])

allowed_failures = {"many references": True}

cmark = CMark(prog=args.program, library_dir=args.library_dir)

def hash_collisions():
    REFMAP_SIZE = 16
    COUNT = 25000

    def badhash(ref):
        h = 0
        for c in ref:
            a = (h << 6) & 0xFFFFFFFF
            b = (h << 16) & 0xFFFFFFFF
            h = ord(c) + a + b - h
            h = h & 0xFFFFFFFF

        return (h % REFMAP_SIZE) == 0

    keys = ("x%d" % i for i in itertools.count())
    collisions = itertools.islice((k for k in keys if badhash(k)), COUNT)
    bad_key = next(collisions)

    document = ''.join("[%s]: /url\n\n[%s]\n\n" % (key, bad_key) for key in collisions)

    return document, re.compile(r"(<p>\[%s]</p>\n){%d}" % (bad_key, COUNT-1))


# list of pairs consisting of input and a regex that must match the output.
pathological = {
    # note - some pythons have limit of 65535 for {num-matches} in re.
    "nested strong emph":
                (("*a **a " * 32500) + "b" + (" a** a*" * 32500),
                 re.compile("(<em>a <strong>a ){32500}b( a</strong> a</em>){32500}")),
    "many emph closers with no openers":
                 (("a_ " * 32500),
                  re.compile("(a[_] ){32499}a_")),
    "many emph openers with no closers":
                 (("_a " * 32500),
                  re.compile("(_a ){32499}_a")),
    "many link closers with no openers":
                 (("a]" * 32500),
                  re.compile("(a\\]){32500}")),
    "many link openers with no closers":
                 (("[a" * 32500),
                  re.compile("(\\[a){32500}")),
    "mismatched openers and closers":
                 (("*a_ " * 25000),
                  re.compile("([*]a[_] ){24999}[*]a_")),
    "issue #389":
                 (("*a " * 20000 + "_a*_ " * 20000),
                  re.compile("(<em>a ){20000}(_a<\\/em>_ ?){20000}")),
    "openers and closers multiple of 3":
                 (("a**b" + ("c* " * 25000)),
                  re.compile("a[*][*]b(c[*] ){24999}c[*]")),
    "link openers and emph closers":
                 (("[ a_" * 25000),
                  re.compile("(\\[ a_){25000}")),
    "pattern [ (]( repeated":
                 (("[ (](" * 40000),
                  re.compile("(\\[ \\(\\]\\(){40000}")),
    "pattern ![[]() repeated":
                 ("![[]()" * 160000,
                  re.compile("(!\\[<a href=\"\"></a>){160000}")),
    "hard link/emph case":
                 ("**x [a*b**c*](d)",
                  re.compile("\\*\\*x <a href=\"d\">a<em>b\\*\\*c</em></a>")),
    "nested brackets":
                 (("[" * 25000) + "a" + ("]" * 25000),
                  re.compile("\\[{25000}a\\]{25000}")),
    "nested block quotes":
                 ((("> " * 25000) + "a"),
                  re.compile("(<blockquote>\n){25000}")),
    "deeply nested lists":
                 ("".join(map(lambda x: ("  " * x + "* a\n"), range(0,500))),
                  re.compile("<ul>\n(<li>a\n<ul>\n){499}<li>a</li>\n</ul>\n(</li>\n</ul>\n){499}")),
    "U+0000 in input":
                 ("abc\u0000de\u0000",
                  re.compile("abc\ufffd?de\ufffd?")),
    "backticks":
                 ("".join(map(lambda x: ("e" + "`" * x), range(1,2500))),
                  re.compile("^<p>[e`]*</p>\n$")),
    "unclosed links A":
                 ("[a](<b" * 30000,
                  re.compile(r"(\[a]\(&lt;b){30000}")),
    "unclosed links B":
                 ("[a](b" * 30000,
                  re.compile(r"(\[a]\(b){30000}")),
    "unclosed <!--":
                 ("</" + "<!--" * 300000,
                  re.compile("\\&lt;\\/(\\&lt;!--){300000}")),
    "empty lines in deeply nested lists":
                 ("- " * 30000 + "x" + "\n" * 30000,
                  re.compile(r"^(<\w+>\n?)+x(</\w+>\n)+$")),
    "empty lines in deeply nested lists in blockquote":
                 ("> " + "- " * 30000 + "x\n" + ">\n" * 30000,
                  re.compile(r"^(<\w+>\n?)+x(</\w+>\n)+$")),
    "emph in deep blockquote":
                 (">" * 100000 + "a*" * 100000,
                  re.compile(r"^(<\w+>\n)+<p>.*</p>\n(</\w+>\n)+$")),
    "reference collisions": hash_collisions()
#    "many references":
#                 ("".join(map(lambda x: ("[" + str(x) + "]: u\n"), range(1,5000 * 16))) + "[0] " * 5000,
#                  re.compile("(\\[0\\] ){4999}"))
    }

pathological_cmark = {
    "nested inlines":
                 ("*" * 20000 + "a" + "*" * 20000,
                  re.compile("^\\*+a\\*+$")),
    }

whitespace_re = re.compile('/s+/')

def run_pathological(q, inp):
    q.put(cmark.to_html(inp))

def run_pathological_cmark(q, inp):
    q.put(cmark.to_commonmark(inp))

def run_tests():
    q = multiprocessing.Queue()
    passed = []
    errored = []
    failed = []
    ignored = []

    print("Testing pathological cases:")
    for description in (*pathological, *pathological_cmark):
        if description in pathological:
            (inp, regex) = pathological[description]
            p = multiprocessing.Process(target=run_pathological,
                      args=(q, inp))
        else:
            (inp, regex) = pathological_cmark[description]
            p = multiprocessing.Process(target=run_pathological_cmark,
                      args=(q, inp))
        p.start()
        try:
            # wait TIMEOUT seconds or until it finishes
            rc, actual, err = q.get(True, TIMEOUT)
            p.join()
            if rc != 0:
                print(description, '[ERRORED (return code %d)]' %rc)
                print(err)
                if description in allowed_failures:
                    ignored.append(description)
                else:
                    errored.append(description)
            elif regex.search(actual):
                print(description, '[PASSED]')
                passed.append(description)
            else:
                print(description, '[FAILED]')
                print(repr(actual[:60]))
                if description in allowed_failures:
                    ignored.append(description)
                else:
                    failed.append(description)
        except queue.Empty:
            p.terminate()
            p.join()
            print(description, '[TIMEOUT]')
            if description in allowed_failures:
                ignored.append(description)
            else:
                errored.append(description)

    print("%d passed, %d failed, %d errored" %
          (len(passed), len(failed), len(errored)))
    if ignored:
        print("Ignoring these allowed failures:")
        for x in ignored:
            print(x)
    if failed or errored:
        exit(1)
    else:
        exit(0)

if __name__ == "__main__":
    run_tests()



#!/usr/bin/env python

# Example for using the shared library from python
# Will work with either python 2 or python 3
# Requires cmark library to be installed

from ctypes import *
import sys
import platform

sysname = platform.system()

if sysname == 'Darwin':
    libname = "libcmark.dylib"
elif sysname == 'Windows':
    libname = "cmark.dll"
else:
    libname = "libcmark.so"
cmark = CDLL(libname)

class cmark_mem(Structure):
    _fields_ = [("calloc", c_void_p),
                ("realloc", c_void_p),
                ("free", CFUNCTYPE(None, c_void_p))]

get_alloc = cmark.cmark_get_default_mem_allocator
get_alloc.restype = POINTER(cmark_mem)
free_func = get_alloc().contents.free

markdown = cmark.cmark_markdown_to_html
markdown.restype = POINTER(c_char)
markdown.argtypes = [c_char_p, c_size_t, c_int]

opts = 0 # defaults

def md2html(text):
    text = text.encode('utf-8')
    cstring = markdown(text, len(text), opts)
    result = string_at(cstring).decode('utf-8')
    free_func(cstring)
    return result

sys.stdout.write(md2html(sys.stdin.read()))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re
import os
import argparse
import sys
import platform
import html
from cmark import CMark

def get_entities():
    regex = r'^{\(unsigned char\*\)"([^"]+)", \{([^}]+)\}'
    with open(os.path.join(os.path.dirname(__file__), '..', 'src', 'entities.inc')) as f:
        code = f.read()
    entities = []
    for entity, utf8 in re.findall(regex, code, re.MULTILINE):
        utf8 = bytes(map(int, utf8.split(", ")[:-1])).decode('utf-8')
        entities.append((entity, utf8))
    return entities

parser = argparse.ArgumentParser(description='Run cmark tests.')
parser.add_argument('--program', dest='program', nargs='?', default=None,
        help='program to test')
parser.add_argument('--library-dir', dest='library_dir', nargs='?',
        default=None, help='directory containing dynamic library')
args = parser.parse_args(sys.argv[1:])

cmark = CMark(prog=args.program, library_dir=args.library_dir)

entities = get_entities()

passed = 0
errored = 0
failed = 0

exceptions = {
    'quot': '&quot;',
    'QUOT': '&quot;',

    # These are broken, but I'm not too worried about them.
    'nvlt': '&lt;⃒',
    'nvgt': '&gt;⃒',
}

print("Testing entities:")
for entity, utf8 in entities:
    [rc, actual, err] = cmark.to_html("&{};".format(entity))
    check = exceptions.get(entity, utf8)

    if rc != 0:
        errored += 1
        print(entity, '[ERRORED (return code {})]'.format(rc))
        print(err)
    elif check in actual:
        # print(entity, '[PASSED]') # omit noisy success output
        passed += 1
    else:
        print(entity, '[FAILED]')
        print(repr(actual))
        failed += 1

print("{} passed, {} failed, {} errored".format(passed, failed, errored))
if failed == 0 and errored == 0:
    exit(0)
else:
    exit(1)
